/**********************************************************************
 *
 *     umatrix.c  -  Routines for matrix manipulation.
 *                        Version 2.0
 *
 * NOTICE: Copyright (C) 2000, 2001 by Robert S. Larkin
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The program author can be contacted by email at boblark@proaxis.com.
 * Alternatively contact the author at
 *     Robert S. (Bob) Larkin
 *     2982 N. W. Acacia Place
 *     Corvallis, OR  97330  USA
 *
 * Note this is not intended to restrict the distribution or copyright
 * character of any material that may have been used as a basis for this work.
 *
 * For more details on the background or use of the program, pay a visit to
 * http://www.proaxis.com/~boblark/dsp10.htm  and,
 * http://www.ao.com/~gnome/
 *
 * Initial write 27 Feb 2000 RSL
 * Corrected variable typecasting 11Jan 2001, RSL  V2.0
 */

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <bios.h>
#include <math.h>
#include <float.h>
#include <alloc.h>
#include <ctype.h>
#include <graphics.h>
#include <string.h>
#include "umatrix.h"

/**************************************************************************
 *
 * UMATRIX.C - Source code for curve fitting & associated  matrix functions
 *
 * fit_curve(...)  -  Function to find the least-squares 0-order curve
 * fit that is put into a column vector cf[ ] that must be o elements or more.
 * Input data y[] a function of x[] and weight w[], all of length n (0, n-1).
 *
 * Returns 1 unless matrix function fails. Then returns 0 with
 * an error summary put into caller supplied global variable mat_err
 * (see below).
 *
 * Support functions:
 * mat_allocate   allocate matrix, return pointer to MAT structure
 * mat_free       free matrix area and MAT structure
 * mat_transpose  transpose a matrix, return pointer to new MAT
 * mat_mult       multiply two matrices, return new MAT
 * mat_invert     invert a square matrix, return inverse MAT
 *
 * Requires a global variable 'unsigned int mat_err;'
 *   Values: Bit 0   '1'   Memory allocation error
 *               1   '2'   Convergence error
 *               2   '4'   Singularity
 *               3   '8'   Invalid matrix structure
 *               4  '16'   Invalid data type
 *               5  '32'   Invert call with non-square matrix
 *               6  '64'   Mult call with invalid matrix sizes
 *
 *************************************************************************/

int fit_curve(double *cf, double *xx, double *yy, double *ww, int n, int o);

/* Local functions: */
struct MAT *mat_allocate(int, int);
int mat_free(struct MAT *);
struct MAT *mat_mult(struct MAT *, struct MAT *);
struct MAT *mat_transpose(struct MAT *);
struct MAT *mat_invert(struct MAT *);

/************************************************************************/

int fit_curve(double *cf, double *xx, double *yy, double *ww, int n, int o)
    {int             i, j;
     double          **z;
     struct MAT      *At, *B,  *At_Ai_At, *Z;
     struct MAT      *Mi, *Mm, *Mt;
     double          **at;

     /* Get storage for weighted powers of x matrix transpose = At */
     if ( (At = mat_allocate(o+1, n)) == 0) return 0;
     at = (double **) At->ptr;

     for(i=0; i<n; i++)
         {
         at[0][i] = ww[i];
         for(j=1;  j<(o+1);  j++)
             at[j][i] = at[j-1][i] * xx[i];
         }

     /* Get storage for Z=weighted y vector */
     if ( (Z=mat_allocate(n, 1)) == 0)  return 0;
     z = (double **) Z->ptr;
     for(i=0;  i<n;  i++) z[i][0] = ww[i] * yy[i];

     /* Break the matrix operations into single steps, so that we
        can check errors as we go, and can free memory at the end of
        the function.    */
     if((Mt=mat_transpose(At)) == 0) return 0;
     if((Mm=mat_mult(At, Mt)) == 0) return 0;
     if((Mi=mat_invert(Mm)) == 0)  return 0;
     if((At_Ai_At=mat_mult(Mi, At)) == 0) return 0;
     if((B = mat_mult(At_Ai_At,Z)) == 0) return 0;

     /* To make global use simpler, fill in a user supplied one-dimensional
        vector with the polynomial coefficients.  */
     for (i=0;  i<=o;  i++)
         cf[i]=((double **)B->ptr)[i][0];

     /* Very important to free up memory allocated:   */
     if ( mat_free(At) & mat_free(Z) &
              mat_free(At_Ai_At) & mat_free(B) &
              mat_free(Mi) & mat_free(Mm) & mat_free(Mt) )
          return 1;
     else
          return 0;
    }

struct MAT* mat_allocate(int rows, int cols)
    {
    int i;
    struct MAT *A;

    A = (struct MAT *) calloc(1,sizeof(struct MAT));
    if(!A)
       {mat_err |= 1; return 0;
       }
    A->rows = rows;
    A->cols = cols;

    A->ptr = (double **)calloc(rows,sizeof(double *));
    if(!A->ptr)
        {mat_err |= 1; return 0;
        }
    for(i = 0 ; i < rows ; i++)
        {
        A->ptr[i] = (double *)calloc(cols,sizeof(double));
        if(!A->ptr[i])
            {mat_err |= 1; return 0;
            }
        }
    return(A);
    }

int mat_free(struct MAT* A)
    {
    int i;
    double **a;

    if(!A || !A->ptr || !A->rows || !A->cols)
        {mat_err |= 8; return 0;
        }
    a = A->ptr;
    for(i=0;  i<A->rows;  i++) free(a[i]);
    free(a);
    free(A);
    return 1;
    }

struct MAT *mat_transpose(struct MAT* A)
    {
    struct MAT *At;
    double **ad,  **adt;
    int i, j;

    At = mat_allocate(A->cols, A->rows);
    ad = (double **)A->ptr;
    adt = (double **)At->ptr;
    for(i = 0 ; i < A->rows ; i++)
        for(j = 0 ; j < A->cols ; j++)
            adt[j][i] = ad[i][j];
    return(At);
    }

struct MAT* mat_mult(struct MAT* A, struct MAT* B)
    {
    int i, j, k;
    double* pa;
    double* pb;
    double* pc;
    struct MAT *C;
    int ar, ac, bc;

    if(B->rows != A->cols)
        {mat_err |= 64; return 0;
        }
    if(!A->ptr)
        {mat_err |= 8; return 0;
        }
    if(!B->ptr)
        {mat_err |= 8; return 0;
        }

    C = mat_allocate(A->rows, B->cols);
    if(C==0)
	{mat_err |= 1; return 0;
	}

    for(i=0;  i < A->rows; i++)
	{
	pc = C->ptr[i];
        pb = B->ptr[0];
        for(j=0;  j < B->cols;  j++)
            {
            pa = A->ptr[i];
            *pc = (*pa++) * (*pb++);
            for(k=1;  k < A->cols;  k++)
                *pc += (*pa++) * B->ptr[k][j];
            pc++;
            }
         }
    return(C);
    }

struct MAT *mat_invert(struct MAT* A)
    {
    struct MAT *Ai;
    double **a;
    double big, pivot_inverse, temp, abs_element;
    int *pivot_flag, *swap_col, *swap_row;
    int i, n, row, col, swap, irow, icol;
    int ii, jj;
    double* pa;
    double* pai;
    double** ax;
    double** aix;

    if(A->rows != A->cols)
        {mat_err |= 32; return 0;
        }
    if(!A->ptr)
        {mat_err |= 8; return 0;
        }
    n = A->rows;
    Ai = mat_allocate(n, n);
    if (Ai==0)
        {mat_err |= 1; return 0;
        }
    ax = A->ptr;
    aix = Ai->ptr;
    for(ii=0; ii<n; ii++)
        {
        pai = aix[ii];
        pa = ax[ii];
        for(jj=0; jj<n; jj++)
            *pai++ = *pa++;
        }
    a = (double **)Ai->ptr;
    pivot_flag = (int *) calloc(n,sizeof(int));
    swap_row = (int *) calloc(n,sizeof(int));
    swap_col = (int *) calloc(n,sizeof(int));
    if(!pivot_flag || !swap_row || !swap_col)
        {
        mat_err |= 1;
        return 0;
        }

    for(i = 0 ; i < n ; i++)
        {
        big = 0.0;
        for(row = 0 ; row < n ; row++)
            {
            if(!pivot_flag[row])
                {
                for(col = 0 ; col < n ; col++)
                    {
                    if(!pivot_flag[col])
                        {
                        abs_element = fabs(a[row][col]);
                        if(abs_element >= big)
                            {
                            big = abs_element;
                            irow = row;
                            icol = col;
                            }
                         }
                    }
                }
            }
        pivot_flag[icol]++;

        if(irow != icol)
            {
            for(col = 0 ; col < n ; col++)
                {
                temp = a[irow][col];
                a[irow][col] = a[icol][col];
                a[icol][col] = temp;
                }
            }
        swap_row[i] = irow;
        swap_col[i] = icol;
        if(fabs(a[icol][icol]) < 1.0e-100)
            {
            mat_err |= 4;
            return 0;
            }
        pivot_inverse = 1.0/a[icol][icol];
        a[icol][icol] = 1.0;        /* pivot = 1 to avoid round off */
        for(col = 0 ; col < n ; col++)
            a[icol][col] = a[icol][col]*pivot_inverse;
        for(row = 0 ; row < n ; row++)
            {
            if(row != icol)
                {
                temp = a[row][icol];
                a[row][icol] = 0.0;
                for(col = 0 ; col < n ; col++)
                    {
                    a[row][col] = a[row][col]-a[icol][col]*temp;
		    }
                }
            }
        }
    for(swap = n-1 ; swap >= 0 ; swap--)
        {
        if(swap_row[swap] != swap_col[swap])
            {
            for(row = 0 ; row < n ; row++)
                {
                temp = a[row][swap_row[swap]];
                a[row][swap_row[swap]] = a[row][swap_col[swap]];
                a[row][swap_col[swap]] = temp;
                }
            }
        }
    free(pivot_flag);
    free(swap_row);
    free(swap_col);

    return(Ai);
    }
